
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.*;

import java.io.*;
import java.net.InetAddress;
import java.net.Socket;
import java.util.ArrayList;

/**
 * This Class is stand for a client side program for a P2P system.
 * This clss is the main class of the client side program.it contains
 * GUI and also is responsible for making socket connections to the server.
 * it also contains an inner class which is responsible for downloading files
 * from other peers.
 *
 * @see Server
 */

public class Client {

    /**
     * This field contains the server IP which is used to make a connection.
     * this field is extracted from Client Configuration file.
     */
    private String serverIP;

    /**
     * This field contains the server port which is used to make a connection.
     * this field is also extracted from Client Configuration file.
     */
    private int serverPort;

    /**
     * This field contains the port which is used by other clients to connect to this client.
     * this field is also extracted from Client Configuration file.
     */
    private int port;

    /**
     * this is the sharing directory which is used in this program.
     * this field is also extracted from Client Configuration file.
     */
    private String path;

    /**
     * Constructs a new client and make connetions and show GUI to users.
     * it also contains an inner class which is used to download requests
     */
    public Client() {

        // Reading information from configuration file.
        try {
            BufferedReader conf = new BufferedReader(new FileReader("CLIENT.cfg"));
            serverIP = conf.readLine();
            serverPort = Integer.parseInt(conf.readLine());
            port = Integer.parseInt(conf.readLine());
            path = conf.readLine();
            conf.close();
        } catch (FileNotFoundException fnfe) {
            System.out.println("Cannot locate input file! " + fnfe.getMessage());
            System.exit(0);
        } catch (IOException ioe) {
            System.out.println("The config file is empty! " + ioe.getMessage());
            System.exit(0);
        }

        // Make a Recieve Query Object to answer incoming queries from server
        ReceiveQuery r = new ReceiveQuery(path, serverPort, port, serverIP);
        r.start();

        // Make a WaitForUploadConnection Object for waiting for upload requests
        WaitForUploadConnection waitForUpdateConnection = new WaitForUploadConnection(port, path);
        waitForUpdateConnection.start();

        // This part is the Designing of the GUI
        final Display display = new Display();
        final Shell mainShell = new Shell(display, SWT.CLOSE | SWT.MIN);
        mainShell.setSize(340, 70);
        mainShell.setText("P2P Client");
        mainShell.setImage(new Image(display, "./Client/k.jpg"));

        final Text searchQuery = new Text(mainShell, SWT.BORDER);
        searchQuery.setBounds(9, 10, 200, 20);
        searchQuery.setTextLimit(30);

        Button search = new Button(mainShell, SWT.PUSH);
        search.setText("Search");
        search.setLocation(215, 10);
        search.setSize(50, 20);

        search.addSelectionListener(new SelectionAdapter() {

            // This Anounymous Inner Class is the action Listener for Search Bottun.
            public void widgetSelected(SelectionEvent e) {
                if (searchQuery.getText().length() <= 0)
                    return;

                final Shell resultShell = new Shell(display, SWT.CLOSE | SWT.MIN);
                resultShell.setSize(280, 290);
                resultShell.setText("Search Results");
                resultShell.setImage(new Image(display, "./Client/k.jpg"));

                final ArrayList hitList = new ArrayList();
                final ArrayList hitAddressList = new ArrayList();
                final ArrayList hitPortList = new ArrayList();

                final List result = new List(resultShell, SWT.SINGLE | SWT.BORDER | SWT.V_SCROLL | SWT.H_SCROLL);
                result.setBounds(9, 10, 255, 210);
                result.setToolTipText("Double Click on Each Item to Download");
                result.addMouseListener(new MouseAdapter() {

                    /**
                     * This is a progressbar object which is used to display download status.
                     */
                    private ProgressBar progressBar = null;

                    /////////////////////////////////////////

                    /**
                     * this method is responsible for downloading requested file from the specified location.
                     *
                     * @param filename this is the name of the file that is requested to be downloaded.
                     * @param address  this is the address of the peer that has the specified file.
                     * @param port     this is the port number of the peer that has the specified file.
                     */
                    private void download(String filename, InetAddress address, int port) {

                        Socket client = null;

                        // Connect to the Client to Download file.
                        try {
                            client = new Socket(address, port);
                        } catch (IOException e) {
                            System.out.println("Client : Cannot Connect to Target Client to get files");
                            return;
                        }

                        ObjectOutputStream output = null;
                        ObjectInputStream input = null;

                        try {
                            output = new ObjectOutputStream(client.getOutputStream());
                            input = new ObjectInputStream(client.getInputStream());
                        } catch (IOException e) {
                            System.out.println("Client : Cannot Make Streams for Client Target");
                            try {
                                client.close();
                            } catch (IOException e1) {
                                System.out.println("Client : Error Closing Download ");
                            }
                            return;
                        }

                        // Opening file for storing file which is downloading.
                        try {
                            File file = new File(path + "Download\\");
                            if (!file.exists()) {
                                file.mkdir();
                            }
                            FileOutputStream outputfile = new FileOutputStream(path + "Download\\" + filename);

                            // sending file name
                            output.writeObject(filename);
                            output.reset();
                            output.flush();
                            Long size = null;
                            // recieving file size.
                            try {
                                size = (Long) input.readObject();
                            } catch (ClassNotFoundException e1) {
                                System.out.println("Error : Socket is corrupted !!");
                                return;
                            }

                            long currentsize = 0;
                            int temp = 0;
                            byte bytes [] = new byte[10000];
                            // recieving file bytes.
                            while (currentsize < size.longValue() - 1) {
                                temp = input.read(bytes);
                                outputfile.write(bytes, 0, temp);
                                currentsize += temp;
                                if (size.longValue() != 0) {
                                    int k = (int) (currentsize * 100 / size.longValue());
                                    progressBar.setSelection(k);
                                    //progressBar.redraw();
                                }
                            }
                        } catch (IOException e) {
                            System.out.println("Client : Error Downloading File");
                            try {
                                output.close();
                                input.close();
                                client.close();
                            } catch (IOException e1) {
                                System.out.println("Client : Error Closing Download ");
                            }
                            return;
                        }

                        try {
                            output.close();
                            input.close();
                            client.close();
                        } catch (IOException e1) {
                            System.out.println("Client : Error Closing Download ");
                        }
                        return;

                    }
                    /////////////////////////////////////////

                    public void mouseDoubleClick(MouseEvent e) {
                        if (result.getSelectionCount() == 0)
                            return;

                        final Shell downloadShell = new Shell(display, SWT.CLOSE | SWT.MIN);
                        downloadShell.setSize(300, 130);
                        downloadShell.setText("Downloading Status");
                        downloadShell.setImage(new Image(display, "./Client/k.jpg"));

                        Label label = new Label(downloadShell, SWT.NONE);
                        label.setText("Downloading: " + result.getSelection()[0]);
                        label.setSize(265, 20);
                        label.setLocation(12, 10);

                        Button ok = new Button(downloadShell, SWT.NONE);
                        ok.setText("Ok");
                        ok.setSize(40, 20);
                        ok.setLocation(130, 70);
                        ok.setEnabled(false);
                        ok.addSelectionListener(new SelectionAdapter() {
                            public void widgetSelected(SelectionEvent e) {
                                downloadShell.dispose();
                            }
                        });

                        progressBar = new ProgressBar(downloadShell, SWT.HORIZONTAL);
                        progressBar.setMinimum(0);
                        progressBar.setMaximum(100);
                        progressBar.setBounds(10, 40, 270, 20);

                        downloadShell.open();

                        download((String) hitList.get(result.getSelectionIndex()), (InetAddress) hitAddressList.get(result.getSelectionIndex()), ((Integer) hitPortList.get(result.getSelectionIndex())).intValue());

                        label.setText("Download Completed !!");
                        ok.setEnabled(true);

                        while (!downloadShell.isDisposed()) {

                            if (!display.readAndDispatch())
                                display.sleep();
                        }
                    }
                });

                Button close = new Button(resultShell, SWT.PUSH);
                close.setText("Close");
                close.setLocation(117, 232);
                close.setSize(50, 20);
                close.addSelectionListener(new SelectionAdapter() {
                    public void widgetSelected(SelectionEvent e) {
                        resultShell.dispose();
                    }
                });

                // Making a Search Object for performing a search.
                Search s = new Search(searchQuery.getText(), serverIP, serverPort, hitList, hitAddressList, hitPortList);
                s.start();
                int count = 0;

                searchQuery.setText("");
                resultShell.open();

                while (!resultShell.isDisposed()) {
                    synchronized (hitList) {
                        while (count < hitList.size())
                            result.add((String) hitList.get(count++));
                    }

                    if (!display.readAndDispatch())
                        display.sleep();
                }
            }
        });

        mainShell.open();

        Button quit = new Button(mainShell, SWT.PUSH);
        quit.setText("Quit");
        quit.setLocation(270, 10);
        quit.setSize(50, 20);

        quit.addSelectionListener(new SelectionAdapter() {
            public void widgetSelected(SelectionEvent e) {
                mainShell.close();
            }
        });
        mainShell.setDefaultButton(search);


        while (!mainShell.isDisposed()) {
            if (!display.readAndDispatch())
                display.sleep();
        }
        // Finalizing the program and closing connections.
        try {
            r.end();
        } catch (Exception e) {
            System.out.println("Error : Ending RecieveQuery ");
        }
        try {
            waitForUpdateConnection.end();
        } catch (Exception e) {
            System.out.println("Error : Ending WaitForUpdateConnection");
        }
        display.close();
    }

    public static void main(String[] args) {
        new Client();
    }
}

