
import java.io.*;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;

/**
 * This class is responsible for commiunication between clients.
 * This consists of a acceptor and dispatcher which detect what must be
 * done to the connection.
 *
 * @author Naeem & Kamyar
 */
public class Server {
    /**
     * The port that server listens to it
     */
    private int port;

    /**
     * List of the queues of the receiuve queries
     */
    private ArrayList queuesList;

    /**
     * server itself
     */
    private ServerSocket server;

    public Server() {
        try {
            /*
            In the first line of the Config file the port to which server
            must listen is placed
            */
            BufferedReader conf = new BufferedReader(new FileReader("SERVER.cfg"));
            port = Integer.parseInt(conf.readLine());
            conf.close();
        } catch (FileNotFoundException fnfe) {
            System.out.println("Cannot locate input file! " + fnfe.getMessage());
            System.exit(0);
        } catch (IOException ioe) {
            System.out.println("The config file is empty! " + ioe.getMessage());
            System.exit(0);
        }

        queuesList = new ArrayList();
    }

    /**
     * Starts the server
     */
    public void startUp() {
        //openning server side socket for waiting for new connections
        try {
            server = new ServerSocket(port);
        } catch (IOException e) {
            System.out.println("Error : Cannot Open Server Side Socket !!");
            return;
        }

        while (true) {
            //waiting for new connections
            try {
                Socket client = server.accept();

                /*
                The blocking thread only accepts connectin and dispatcher
                Detect to which thread it must be passed
                */
                ConnectionDetector connectiondetector = new ConnectionDetector(client);
                connectiondetector.start();
            } catch (IOException e) {
                System.out.println("Error : Cannot Accept new Connections !!");
                return;
            }
        }
    }

    /**
     * Detects if this connection is for search or for supporting material
     * for server.
     */
    private class ConnectionDetector extends Thread {
        /**
         * The other end is client
         */

        Socket client;

        /**
         * Object input stream over client
         */
        ObjectInputStream input;

        /**
         * Object output stream over client
         */
        ObjectOutputStream output;

        public ConnectionDetector(Socket client) {
            this.client = client;
        }

        /**
         * @see Thread#run()
         */
        public void run() {
            try {
                output = new ObjectOutputStream(client.getOutputStream());
                input = new ObjectInputStream(client.getInputStream());
            } catch (IOException e) {
                System.out.println("Error : Cannot Stablish Stream Over Connections !!");
                return;
            }

            try {
                String type = null;

                try {
                    //Reading from client to understang which kind of connectin
                    type = (String) input.readObject();
                } catch (IOException e) {
                    return;
                } catch (ClassNotFoundException e) {
                    return;
                }

                Thread newthread = null;

                if (type.equals("Hello New Client")) {
                    newthread = new SendQuery(client, queuesList, input, output);
                    System.out.println("Making new Send Query");
                } else if (type.equals("Hello New Search")) {
                    newthread = new WaitForQuery(client, queuesList, input, output);
                    System.out.println("Making new Wait for Query");
                } else {
                    System.out.println("Error : Invalid Connection !!");
                    System.out.println("Server : Closinq Socket for bad Connection");
                    client.close();
                    return;
                }

                newthread.start();

            } catch (IOException e) {
                System.out.println("Error : Reading from client socket !!");
                return;
            }
        }
    }

    public static void main(String[] args) {
        Server s = new Server();
        s.startUp();
    }
}
